from typing import List
import logging

import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from PIL import Image
import simplejpeg

# not a strict requirement unless loading fits
try:
    from astropy.io import fits
except ImportError:
    pass



# https://pytorch.org/tutorials/beginner/basics/data_tutorial.html
class GalaxyDataset(Dataset):
    def __init__(self, catalog: pd.DataFrame, label_cols=None, transform=None, target_transform=None):
        """
        Create custom PyTorch Dataset using catalog of galaxy images

        Catalog and images should already be downloaded.
        Use methods under galaxy_datasets.prepared_datasets e.g. prepared_datasets.gz2()
        
        Note that if you want the canonical datasets (i.e. using the standard train/test split)
        you can use the galaxy_datasets.pytorch.datasets instead of this class e.g. 

        ```
        datasets.GZ2Dataset(
            root=os.path.join(some_dir, 'gz2'),
            download=True
        )
        ```
        
        Reads a catalog of jpeg galaxy images
        if transform is from albumentations, datamodule should know about the transform including 
        catalog should be split already
        should have correct image locations under file_loc

        Args:
            catalog (pd.DataFrame): with images under 'file_loc' and labels under label_cols
            label_cols (List): columns in catalog with labels
            transform (callable, optional): See Pytorch Datasets. Defaults to None.
            target_transform (callable, optional): See Pytorch Datasets. Defaults to None.
        """
        # internal catalog has id_str as index for query speed
        # (dataset.catalog still returns the int-indexed catalog via the catalog method below)
        # self._catalog = catalog.copy().set_index('id_str', verify_integrity=True)
        self.catalog = catalog
        
        self.label_cols = label_cols
        self.transform = transform
        self.target_transform = target_transform


    def __len__(self) -> int:
        return len(self.catalog)


    def __getitem__(self, idx: int):
        #the index is id_str so can use that for quick search on 1M+ catalo
        # galaxy = self._catalog.loc[idx]
        galaxy = self.catalog.iloc[idx]

        # load the image into memory
        image_loc = galaxy['file_loc']
        try:
            image = load_img_file(image_loc)
            # HWC PIL image
            # logging.info((image.shape, torch.max(image), image.dtype, label))  # always 0-255 uint8
        except Exception as e:
            logging.critical('Cannot load {}'.format(image_loc))
            raise e
    
        if self.transform:
            image = self.transform(image)
            # now a CHW tensor, which torchvision wants
            # logging.info((image.shape, torch.max(image), image.dtype, label))  #  should be 0-1 float
        # else:
        #     raise ValueError('Please specify a transform')
        #     # TODO could use some kind of default transform for torch-friendly image, if needed

        if self.label_cols is None:
            return image
        else:
            # load the labels. If no self.label_cols, will 
            label = get_galaxy_label(galaxy, self.label_cols)

            if self.target_transform:
                label = self.target_transform(label)

            
            return image, label

def load_img_file(loc):
    # wrapper around specific loaders below
    # could be more performant with a dict of {format:loader} but doubt significant
    if loc.endswith('png'):
        return load_png_file(loc)
    elif loc.endswith('jpg'):
        return load_jpg_file(loc)
    elif loc.endswith('jpeg'):
        return load_jpg_file(loc)
    elif loc.endswith('fits'):
        return load_fits_file(loc)  # careful, these often need a transform to have reasonable dynamic range
    else:
        raise ValueError('File format of {} not recognised - should be jpeg|jpg (preferred) or png'.format(loc))


def load_jpg_file(loc):
    with open(loc, 'rb') as f:
        return Image.fromarray(decode_jpeg(f.read()))  # HWC PIL image via simplejpeg

def load_png_file(loc):
    return Image.open(loc) # HWC

def load_fits_file(loc):  
    x = fits.open(loc)[0].data.astype(np.float32)
    # assumes single channel - add channel dimension
    return np.expand_dims(x, axis=2)  # HWC

def decode_jpeg(encoded_bytes):
    return simplejpeg.decode_jpeg(encoded_bytes, fastdct=True, fastupsample=True)


def get_galaxy_label(galaxy: pd.Series, label_cols: List) -> np.ndarray:
    # pytorch 1.12 is happy with float32 for both dirichlet and cross-entropy
    return galaxy[label_cols].astype(np.float32).values.squeeze()  # squeeze for if there's one label_col


if __name__ == '__main__':

    # lazy test/example
    import glob
    mixed_file_paths = glob.glob('tests/test_data/png_jpg_mix/*')
    assert len(mixed_file_paths) > 0
    data = {
        'file_loc': mixed_file_paths, 
        'id_str': [str(x) for x in np.arange(len(mixed_file_paths))]
    }
    df = pd.DataFrame(data)

    dataset = GalaxyDataset(catalog=df)
    for im in dataset:
        im = np.array(im)  # returns PIL.Image, if not given label_cols or transform
        print(im.shape, im.mean(), im.min(), im.max())



from galaxy_datasets.shared.gz_candels import gz_candels
from galaxy_datasets.shared.gz_decals_5 import gz_decals_5
from galaxy_datasets.shared.gz_hubble import gz_hubble
from galaxy_datasets.shared.gz2 import gz2
from galaxy_datasets.shared.demo_rings import demo_rings
from galaxy_datasets.shared.tidal import tidal

from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if INTERNAL_URLS_EXIST:
    from galaxy_datasets.shared.gz_desi import gz_desi
    from galaxy_datasets.shared.gz_rings import gz_rings



import os
from functools import partial
from typing import List
import logging

import pandas as pd
import tensorflow as tf


def get_image_dataset(
    image_paths: List,
    labels=None,
    check_valid_paths=True,
    # new preprocessing options now here
    requested_img_size=None,
    greyscale=False,
    permute_channels=False,
    include_id_str=False) -> tf.data.Dataset:
    """
    Load images in a folder as a tf.data dataset
    Supports jpeg (note the e) and png

    Args:
        image_paths (list): list of image paths to load
        requested_img_size (int or None): e.g. 256 for 256x256x3 image. Assumed square. Will resize if size on disk != this. No resizing if None.
        labels (list or None): If not None, include labels in dataset (see Returns). Must be equal length to image_paths. Defaults to None.
        greyscale (bool): 
        permute_channels (bool):

    Raises:
        FileNotFoundError: at least one path does not match an existing file

    Returns:
        tf.data.Dataset: yielding batches with 'image' key for image array, 'id_str' for the image path, and 'label' if ``labels`` was provided.
    """
    
    assert len(image_paths) > 0
    if isinstance(image_paths, pd.Series):
        image_paths = list(image_paths)
    example_path = image_paths[0]
    assert isinstance(example_path, str)
    logging.info('Image paths to load as dataset: {}'.format(len(image_paths)))

    if check_valid_paths:
        logging.info('Checking if all paths are valid')
        missing_paths = [path for path in image_paths if not os.path.isfile(path)]
        if missing_paths:
            raise FileNotFoundError(f'Missing {len(missing_paths)} images e.g. {missing_paths[0]}')
        logging.info('All paths exist')
    else:
        logging.warning('Skipping valid path check')

    file_format = example_path.split('.')[-1]  # assume all same format

    # load a single image to check the shape
    test_image = load_image_as_element(example_path)['image']
    size_on_disk = test_image.numpy().shape[0]  # x dimension (XYC convention, not yet batched)
    if requested_img_size is None:
        logging.info('No specific image size requested on load - not resizing (until augmentations')
        resize_size = None
    elif size_on_disk == requested_img_size:
        logging.info('Image size on disk matches requested_img_size of {}, skipping resizing'.format(requested_img_size))  # x dimension of first image, first y index, first channel
        resize_size = None
    else:
        logging.warning('Resizing images from disk size {} to requested size {}'.format(size_on_disk, requested_img_size))
        resize_size = requested_img_size

    path_ds = tf.data.Dataset.from_tensor_slices([str(path) for path in image_paths])

    # will yield elements of {'image': image, 'id_str': path}
    image_ds = path_ds.map(lambda x: load_image_as_element(x, mode=file_format), num_parallel_calls=tf.data.AUTOTUNE)  # keep determinstic though

    # applies preprocessing transforms to x['image']: greyscaling/permuting, resizing
    # TODO could join these into the above function for speed, if needed
    image_ds = image_ds.map(lambda x: preprocess_image_in_element(x, resize_size=resize_size, greyscale=greyscale, permute_channels=permute_channels), num_parallel_calls=tf.data.AUTOTUNE)

    if labels is not None:
        assert len(labels) == len(image_paths)

        # if isinstance(labels[0], dict):
        #     # assume list of dicts, each representing one datapoint e.g. [{'feat_a': 1, 'feat_b': 2}, {'feat_a': 12, 'feat_b': 42}]
        #     # reshape to columnlike dict e.g. {'feat_a': [1, 12], 'feat_b: [2, 42]} because that's what tf supports
        #     # (could pass this directly, but inputs tend to be easier to handle row-wise for e.g. shuffling etc)
        #     label_ds = tf.data.Dataset.from_tensor_slices(pd.DataFrame.from_dict(labels).to_dict(orient="list"))
        # else:
            # make it a dict anyway for consistency, keyed by 'label' instead of e.g. 'feat_a', 'feat_b'
        label_ds = tf.data.Dataset.from_tensor_slices({'label': labels})

        # label_dict is {'label': (256)} or {'feat_a': (256), 'feat_b': (256)}
        # image_dict is {'id_str': some_id 'image': (image)}
        # merge the two dicts to create {'id_str': ..., 'image': ..., 'feat_a': ..., 'feat_b': ...}
        image_ds = tf.data.Dataset.zip((image_ds, label_ds)).map(lambda image_dict, label_dict: {**image_dict, **label_dict})
        # now yields {'image': , 'id_str': , 'label': } batched dicts

    # now choose how to unpack
    # (yes, I could probably do this faster/in fewer steps, but this is very readable)
    if include_id_str:
        if labels is not None:
            image_ds = image_ds.map(lambda x: (x['image'], x['label'], x['id_str']))
        else:
            image_ds = image_ds.map(lambda x: (x['image'], x['id_str']))
    else:
        assert labels is not None
        image_ds = image_ds.map(lambda x: (x['image'], x['label']))

    return image_ds


# https://stackoverflow.com/questions/62544528/tensorflow-decodejpeg-expected-image-jpeg-png-or-gif-got-unknown-format-st?rq=1
def load_image_as_element(loc, mode='png'):
    """
    Load an image file from disk to memory.
    *Recently changed to return 0-1 floats not 0-255 floats*

    Args:
        loc (str): Path to image on disk. Includes format e.g. .png.
        mode (str, optional): Image format. Defaults to 'png'.

    Raises:
        ValueError: mode is neither png nor jpeg.

    Returns:
        dict: like {'image': float32 np.ndarray from 0. to 1., 'id_str': ``loc``}
    """
    # values will be 0-255, does not normalise. Happens in preprocessing instead.
    # specify mode explicitly to avoid graph tracing issues
    image = tf.io.read_file(loc)
    if mode == 'png':
        image = tf.image.decode_png(image)
    # rename jpg to jpeg or validation checks in decode_jpg will fail
    # might fail, let's see
    elif (mode == 'jpg') or (mode == 'jpeg'):  
        # TODO also allow jpg
        image = tf.image.decode_jpeg(image)
    else:
        raise ValueError(f'Image filetype mode {mode} not recognised')

    converted_image = tf.cast(image, tf.float32) / 255.  # floats in 0-1 range

    return {'image': converted_image, 'id_str': loc}  # using the file paths as identifiers


def preprocess_image_in_element(element, resize_size=None, greyscale=False, permute_channels=False):
    # now applies all preprocessing

    # element['image'] must be 0-1 floats, not 0-255 ints, or clipping will ruin
    image, id_str = element['image'], element['id_str']  # unpack from dict

    if greyscale:
        # new channel dimension of 1
        image = tf.reduce_mean(input_tensor=image, axis=-1, keepdims=True)
        # assert images.shape[1] == input_size
        # assert images.shape[2] == input_size
        # assert images.shape[3] == 1
        # tf.summary.image('b_make_greyscale', images)
    else:
        if permute_channels:
            image = tf.map_fn(permute_channels, image)  # map to each image in element
        else:
            image = tf.identity(image)
            
    if resize_size is not None:
        image = resize_image_with_tf(image , size=resize_size)   # initial size = after resize from image on disk (e.g. 424 for GZ pngs) but before crop/zoom
        image = tf.clip_by_value(image, 0., 1.)  # resizing can cause slight change in min/max

    return {'image': image, 'id_str': id_str}  # pack back into dict


def resize_image_with_tf(image, size):
    # May cause values outside 0-255 margins
    # allows for image dimension, or no image dimension
    # May be slow. Ideally, resize the images beforehand on disk (or save as TFRecord, see make_shards.py and tfrecord_datasets.py)
    return tf.image.resize(image, (size, size), method=tf.image.ResizeMethod.LANCZOS3, antialias=True)


def permute_channels(im):
    assert tf.shape(im)[-1] == 3
    # tf.random.shuffle shuffles 0th dimension, so need to temporarily swap channel to 0th, shuffle, and swap back
    return tf.transpose(tf.random.shuffle(tf.transpose(im, perm=[2, 1, 0])), perm=[2, 1, 0])


def add_transforms_to_dataset(dataset, transforms):
    # only works if dataset yields either (images, labels) or (images, id_str), not (images, label_str, id_str)
    # TODO make work with extra element - a little fiddly due to unpacking in graph mode, though

    # use closure to pass in transforms (which cannot be passed via inp as not tensor-like)
    def process_data(images, labels):
        aug_imgs = tf.numpy_function(func=aug_fn, inp=[images], Tout=tf.float32)
        return aug_imgs, labels
        # return (aug_imgs) + args  # equivalent to aug_imgs, *args, but works prior to Python 3.8

    def aug_fn(image):
        # albumentations.transforms expects input like image=x
        # and returns {'image': image} dict
        return transforms(image=image)["image"]

    # TODO specify output_size

    return dataset.map(
        partial(process_data),
        num_parallel_calls=tf.data.experimental.AUTOTUNE
    )




from galaxy_datasets.shared.gz2 import gz2
from galaxy_datasets.shared.gz_candels import gz_candels
from galaxy_datasets.shared.gz_decals_5 import gz_decals_5
from galaxy_datasets.shared.gz_hubble import gz_hubble
from galaxy_datasets.shared.tidal import tidal
from galaxy_datasets.shared.demo_rings import demo_rings

from galaxy_datasets import check_internal_urls
if check_internal_urls.INTERNAL_URLS_EXIST:
    from galaxy_datasets.shared.gz_desi import gz_desi
    from galaxy_datasets.shared.gz_rings import gz_rings



import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata


def gz2(root, train, download):
    logging.info('Setting up gz2 dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/vu77e3sh2s5c250/gz2_train_catalog.parquet', 'd5507a9332c75fd84a7f2562567da36b'),  # the train catalog
        ('https://dl.dropboxusercontent.com/s/t8eh6f3oupndpl3/gz2_test_catalog.parquet', 'ac1fba88d0e8a95ee4f4eef79ea03063'),  # the test catalog
        ('https://dl.dropboxusercontent.com/s/5so7yof2afe761p/images_gz2.tar.gz', 'e3eab2fec57a6a60577236b9e0a6913d')  # the images
    ]
    images_to_spotcheck = ['587722/587722981741363294.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()


    # useful_columns = label_cols + ['filename']
    useful_columns = None  # quite a variety of possiblities - load all of them
    if train:
        train_catalog_loc = os.path.join(root, 'gz2_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'gz2_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(
        lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']),
        axis=1
    )

    label_cols = label_metadata.gz2_ortho_label_cols  # default, but you can ignore
    logging.info('gz2 dataset ready')
    return catalog, label_cols




import os
import logging

import pandas as pd

from zoobot.shared import label_metadata

from galaxy_datasets.shared import download_utils, label_metadata
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls


def gz_hubble_euclidized(root, train, download):
    logging.info('Setting up gz_hubble_euclidized dataset')
    resources = [
        (internal_urls.hubble_euclidised_train_catalog, '99a92bb1abd47966f77a7b04e4e310ac'),  # train catalog
        (internal_urls.hubble_euclidised_test_catalog, '4f628f5e7d3282fbfa7ed4b0d7d91b11'),  # test catalog
        (internal_urls.hubble_euclidised_images, 'baee3a003631be238bb9cd6650cf03bf')
    ]
    images_to_spotcheck = ['20000011.jpg', '20164189.jpg', '203191082.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    label_cols = label_metadata.hubble_ortho_label_cols

    useful_columns = label_cols + ['filename']
    if train:
        train_catalog_loc = os.path.join(root, 'catalog_euclid_train.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'catalog_euclid_test.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['filename']), axis=1)  # no subfolder

    logging.info('gz_hubble_euclidized dataset ready')
    return catalog, label_cols


if __name__ == '__main__':


    gz_hubble_euclidized(
        root='/nvme1/scratch/walml/repos/galaxy-datasets/roots/gz_hubble_euclidised',
        train=True,
        download=True
        )


import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls


def gz_rings(root, train, download):
    logging.info('Setting up gz_rings dataset')
    resources = [
        (internal_urls.rings_train_catalog, 'e2fb6b2bca45cd7f1c58f5b4089a5976'),  # train catalog
        (internal_urls.rings_test_catalog, '6e3f362a6e19ecd02675eaa48f6727f0'),  # test catalog
        (internal_urls.rings_images, 'd0950250436a05ce88de747e6af825b6')  # the images
    ]
    images_to_spotcheck = []

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    # useful_columns = label_cols + ['subfolder', 'filename']
    useful_columns = None
    if train:
        train_catalog_loc = os.path.join(root, 'rings_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'rings_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    label_cols = label_metadata.rings_label_cols

    logging.info('gz_rings dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    gz_rings(root='/Users/user/repos/galaxy-datasets/roots/gz_rings', train=True, download=True)


import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata


def gz_hubble(root, train, download):
    logging.info('Setting up gz_hubble dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/xnktj9hq6xig0a7/hubble_ortho_train_catalog.parquet', 'c6cb821f7ebefb583dc74488cf7bfc5f'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/1g9lwih9944sys8/hubble_ortho_test_catalog.parquet', '05e01ed822b34400f32977280eebec87'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/k9xco1mtp8bw60v/hubble_images.tar.gz', None)  # the images 'b6fed2463bb2d17ddb8302f6b060534a'
    ]
    images_to_spotcheck = ['10000325.jpg', '20163083.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    # label_cols = [question + answer for question, answers in hubble_ortho_pairs.items() for answer in answers]  # defined below, globally in this script (for imports elsewhere)
    label_cols = label_metadata.hubble_ortho_label_cols  # see below

    useful_columns = label_cols + ['filename']
    if train:
        train_catalog_loc = os.path.join(root, 'hubble_ortho_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'hubble_ortho_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['filename']), axis=1)  # no subfolder

    logging.info('gz_hubble dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    hubble_cols = [x.replace('-hubble', '') for x in label_metadata.hubble_ortho_label_cols]
    # print('\n'.join(cols))

    decals_cols = [x for x in label_metadata.decals_label_cols]
    # print('\n'.join(cols))

    all_cols = list(set(hubble_cols).union(decals_cols))
    all_cols.sort()
    print('\n'.join(all_cols))
